/*******************************************************************************
*                                  sx127x Driver
*                       ----------------------------
*                       innovating embedded platform
*
* Copyright (c) 2001-2017 Guangzhou ZHIYUAN Electronics Stock Co., Ltd.
* All rights reserved.
*
* Contact information:
* web site:    http://www.zlg.cn/
* e-mail:      ametal.support@zlg.cn
*******************************************************************************/

/**
 * \file
 * \brief  sx127x Driver
 *
 * \internal
 * \par Modification history
 * - 1.00 17-11-22  sup, first implementation.
 * \endinternal
 */
 
#include "stdint.h"
#include "math.h"
#include "stdbool.h"
#include "string.h"
#include "radio.h"
#include "sx127x_lora.h"
#include "sx127x_regs_lora.h"


/**
 * \brief ƵƵٽ
 */
#define __LORA_RF_MID_BAND_THRESH               525000000

/**
 *  RSSI Ҫõĳ (RSSIƫֵ)
 */
#define __LORA_RSSI_OFFSET_LF                   -164
#define __LORA_RSSI_OFFSET_HF                   -157

#define SM_STATE_SET(p_this, state)   ((p_this)->run_state.sm_state = state)
#define SM_STATE_GET(p_this)          ((p_this)->run_state.sm_state)
#define CAD_STATE_SET(p_this, state)  ((p_this)->run_state.cad_state |= state)
#define CAD_STATE_GET(p_this)         ((p_this)->run_state.cad_state)
#define CAD_STATE_CLR(p_this)         ((p_this)->run_state.cad_state = 0)


const double RssiOffsetLF[] =
{   // These values need to be specify in the Lab
    -155.0,
    -155.0,
    -155.0,
    -155.0,
    -155.0,
    -155.0,
    -155.0,
    -155.0,
    -155.0,
    -155.0,
};

const int16_t RssiOffsetHF[] =
{   // These values need to be specify in the Lab
    -150.0,
    -150.0,
    -150.0,
    -150.0,
    -150.0,
    -150.0,
    -150.0,
    -150.0,
    -150.0,
    -150.0,
};



int __sx127x_lora_set_op_mode(radio_sx127x_dev_t *p_this, uint8_t opMode);
int __sx127x_lora_set_rf_frequency (radio_sx127x_dev_t *p_this, uint32_t freq);
int __sx127x_lora_set_payload_length (radio_sx127x_dev_t *p_this, uint8_t value);


static void __delay_ms (radio_sx127x_dev_t  *p_this, uint32_t ms)
{
    p_this->p_delay->pfn_delay_ms(ms);
}




static void __delay_us (radio_sx127x_dev_t  *p_this, uint32_t us)
{
    p_this->p_delay->pfn_delay_us(us);
}



static uint8_t __sx127x_dio0_level_read (radio_sx127x_dev_t *p_this)
{
    return p_this->p_gpio->pfn_dio0_pin_read();
}



static void __send_byte (radio_sx127x_dev_t *p_this, uint8_t byte)
{
    p_this->p_spi->pfn_spi_write_byte(byte);
}




static uint8_t __recv_byte (radio_sx127x_dev_t *p_this)
{
    return p_this->p_spi->pfn_spi_read_byte();
}




static void __sx127x_write_buffer (radio_sx127x_dev_t *p_this,
                                   uint8_t             addr, 
                                   uint8_t            *buf, 
                                   uint8_t             size)
{
    uint8_t i;

    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, addr | 0x80);
    for (i = 0; i < size; i++) {
        __send_byte(p_this, buf[i]);
    }
    
    p_this->p_gpio->pfn_sel_pin_set(1);
}




static void __sx127x_read_buffer (radio_sx127x_dev_t *p_this,
                                  uint8_t             addr, 
                                  uint8_t            *buf, 
                                  uint8_t             size)
{
    uint8_t i;

    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, addr & 0x7F);

    for (i = 0; i < size; i++ ) {
        buf[i] = __recv_byte(p_this);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
}



static void __sx127x_read_fifo (radio_sx127x_dev_t *p_this, uint8_t *buffer, uint8_t size)
{
    __sx127x_read_buffer(p_this, 0, buffer, size);
}



static void __sx127x_write_fifo (radio_sx127x_dev_t *p_this, uint8_t *buffer, uint8_t size)
{
    __sx127x_write_buffer(p_this, 0, buffer, size);
}



static int __sx127x_reset (radio_sx127x_dev_t *p_this)
{
    p_this->p_gpio->pfn_reset_pin_set(0);
    __delay_ms(p_this, 2);
    p_this->p_gpio->pfn_reset_pin_set(1);
    __delay_ms(p_this, 10);
    
    return 0;
}



static void __sx127x_write_reg (radio_sx127x_dev_t *p_this,
                                    uint8_t             addr,
                                    uint8_t             data)
{
    __sx127x_write_buffer(p_this, addr, &data, 1);
}



static uint8_t __sx127x_read_reg (radio_sx127x_dev_t *p_this,
                                  uint8_t             addr)
{
    uint8_t reg;

    __sx127x_read_buffer(p_this, addr, &reg, 1);
    return reg;
}



static void __freq_hop_set (radio_sx127x_dev_t *p_this)
{
    uint8_t hop_period_reg;
    uint8_t hop_channel;
    
    if (p_this->p_config->p_lora_param->freq_hop_on == true) {
        hop_period_reg = p_this->p_config->p_lora_param->hop_period;
        hop_channel    = __sx127x_read_reg(p_this, REG_LR_HOPCHANNEL) & RFLR_HOPCHANNEL_CHANNEL_MASK;
        __sx127x_lora_set_rf_frequency(p_this, p_this->p_config->p_lora_param->p_hop_tab[hop_channel % 
                                                            p_this->p_config->p_lora_param->hop_tab_len]);
    } else {
        hop_period_reg = 255;
    }

    __sx127x_write_reg(p_this, REG_LR_HOPPERIOD, hop_period_reg);
}
    
    
    
static int __sx127x_lora_rx_state_enter (radio_sx127x_dev_t *p_this)
{
    uint8_t reg, reg_val0, reg_val1;
    uint8_t fifo_addr_reg;
    
    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_STANDBY); 

    /* ŵ */
    if (p_this->p_config->p_lora_param->bandwidth < 9) {
        reg = __sx127x_read_reg(p_this, RFLR_DETECTOPTIMIZE);
        __sx127x_write_reg(p_this, RFLR_DETECTOPTIMIZE, reg & 0x7F);
        __sx127x_write_reg(p_this, RFLR_TEST30, 0x00);
        switch (p_this->p_config->p_lora_param->bandwidth) {

        case 0: // 7.8 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x48);
			__sx127x_lora_set_rf_frequency(p_this, 7.81e3);
            break;

        case 1: // 10.4 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x44);
            __sx127x_lora_set_rf_frequency(p_this,p_this->p_config->p_lora_param->frequency + 7.81e3);
            break;

        case 2: // 15.6 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x44);
			__sx127x_lora_set_rf_frequency(p_this,p_this->p_config->p_lora_param->frequency + 15.62e3);
            break;

        case 3: // 20.8 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x44);
			__sx127x_lora_set_rf_frequency(p_this,p_this->p_config->p_lora_param->frequency + 20.83e3);
            break;

        case 4: // 31.2 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x44);
            __sx127x_lora_set_rf_frequency(p_this,p_this->p_config->p_lora_param->frequency + 31.25e3);
            break;

        case 5: // 41.4 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x44);
            __sx127x_lora_set_rf_frequency(p_this,p_this->p_config->p_lora_param->frequency + 41.67e3);
            break;

        case 6: // 62.5 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x40);
            break;

        case 7: // 125 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x40);
            break;

        case 8: // 250 kHz
            __sx127x_write_reg(p_this, RFLR_TEST2F, 0x40);
            break;
        }
    } else {
       // __sx127x_write_reg(p_this, RFLR_TEST2F, 0x3C);
        reg = __sx127x_read_reg(p_this, RFLR_DETECTOPTIMIZE);
        __sx127x_write_reg(p_this, RFLR_DETECTOPTIMIZE, reg | 0x80);
    }

    /* ж  */
    __sx127x_write_reg(p_this, 
                       REG_LR_IRQFLAGSMASK, 
                       //RFLR_IRQFLAGS_RXTIMEOUT |  
                       //RFLR_IRQFLAGS_RXDONE |   
                       //RFLR_IRQFLAGS_PAYLOADCRCERROR |
                       RFLR_IRQFLAGS_VALIDHEADER |
                       RFLR_IRQFLAGS_TXDONE |
                       RFLR_IRQFLAGS_CADDONE |
                       //RFLR_IRQFLAGS_FHSSCHANGEDCHANNEL |
                       RFLR_IRQFLAGS_CADDETECTED);

    /* Ƶ  */
    __freq_hop_set(p_this);
    
    /* Payload */
    __sx127x_lora_set_payload_length(p_this, p_this->p_config->p_lora_param->payload_len);

    /* жϽӳ */
    __sx127x_write_reg(p_this, 
                       REG_LR_DIOMAPPING1, 
                       RFLR_DIOMAPPING1_DIO0_00 |  /* RxDone */
                       RFLR_DIOMAPPING1_DIO1_00 |  /* RxTimeOut */
                       RFLR_DIOMAPPING1_DIO2_01 |  /* Fhss */
                       RFLR_DIOMAPPING1_DIO3_11);  /* -- */
                       
    __sx127x_write_reg(p_this, 
                       REG_LR_DIOMAPPING2, 
                       RFLR_DIOMAPPING2_DIO4_10 |  /* PLL_LOCK */
                       RFLR_DIOMAPPING2_DIO5_11);  /* CLK_OUT */       

    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
    
    if (p_this->p_config->p_lora_param->rx_mode == SINGLE_RECV_MODE) {   /* νģʽ */
        __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_RECEIVER_SINGLE);
    } else {                                                             /* ģʽ */
        fifo_addr_reg = __sx127x_read_reg(p_this, REG_LR_FIFORXBASEADDR);
        __sx127x_write_reg(p_this, REG_LR_FIFOADDRPTR, fifo_addr_reg);   /* SPIFIFOĵַ */
        __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_RECEIVER);
    }
    
    return RADIO_RET_OK;
}



static int __sx127x_lora_set_op_mode(radio_sx127x_dev_t *p_this, uint8_t mode)
{
    static uint8_t mode_prev = RFLR_OPMODE_STANDBY;
    static bool    ant_switch_tx_on_prev = true;
    bool           ant_switch_tx_on = false;
    uint8_t        mode_reg;

    mode_reg   = __sx127x_read_reg(p_this, REG_LR_OPMODE);
    mode_prev  = mode_reg & ~RFLR_OPMODE_MASK;

    if (mode != mode_prev) {
        if (mode == RFLR_OPMODE_TRANSMITTER) {
            ant_switch_tx_on = true;
        } else {
            ant_switch_tx_on = false;
        }

        if (ant_switch_tx_on != ant_switch_tx_on_prev) {
            ant_switch_tx_on_prev = ant_switch_tx_on;
        }
        mode_reg = (mode_reg & RFLR_OPMODE_MASK) | mode;

        __sx127x_write_reg(p_this, REG_LR_OPMODE, mode_reg);
    }
    
    return RADIO_RET_OK;
}



static void __sx127x_set_lora_on (radio_sx127x_dev_t *p_this, bool enable)
{
    uint8_t mode_reg;
    
    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_SLEEP);

    mode_reg = __sx127x_read_reg(p_this, REG_LR_OPMODE);
    mode_reg = (mode_reg & RFLR_OPMODE_LONGRANGEMODE_MASK) | RFLR_OPMODE_LONGRANGEMODE_ON;
    __sx127x_write_reg(p_this, REG_LR_OPMODE, mode_reg);

    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_STANDBY);

    __sx127x_write_reg(p_this, 
                       REG_LR_DIOMAPPING1, 
                       RFLR_DIOMAPPING1_DIO0_00 | /* RxDone    */
                       RFLR_DIOMAPPING1_DIO1_00 | /* RxTimeout */
                       RFLR_DIOMAPPING1_DIO2_00 | /* FhssChangeChannel */
                       RFLR_DIOMAPPING1_DIO3_00); /* CadDone */

    __sx127x_write_reg(p_this, 
                       REG_LR_DIOMAPPING2,
                       RFLR_DIOMAPPING2_DIO4_00 | /* CadDetected */
                       RFLR_DIOMAPPING2_DIO5_00); /* ModeReady */
}



static void __sx127x_lora_set_defaults (radio_sx127x_dev_t *p_this)
{
    uint8_t version_reg;

    version_reg = __sx127x_read_reg(p_this, REG_LR_VERSION);
}



static int __sx127x_lora_set_rf_frequency (radio_sx127x_dev_t *p_this, uint32_t freq)
{
    uint8_t freq_reg[3];
    
    p_this->p_config->p_lora_param->frequency = freq;

    freq = (uint32_t)((double)freq / (double)FREQ_STEP);
    freq_reg[0] = (uint8_t)((freq >> 16) & 0xFF);
    freq_reg[1] = (uint8_t)((freq >> 8) & 0xFF);
    freq_reg[2] = (uint8_t)(freq & 0xFF);
    __sx127x_write_buffer(p_this, REG_LR_FRFMSB, freq_reg, 3);
    
    return RADIO_RET_OK;
}



static void __sx127x_lora_set_nb_trig_peaks (radio_sx127x_dev_t *p_this, uint8_t value)
{
    uint8_t test_reserved_reg;
    
    test_reserved_reg = (__sx127x_read_reg(p_this, 0x31) & 0xF8) | value;
    __sx127x_write_reg(p_this, 0x31, test_reserved_reg);
}



static void __sx127x_lora_set_spreading_factor (radio_sx127x_dev_t *p_this, uint8_t factor)
{
    uint8_t modem_config2_reg;
    
    if (factor > 12) {
        factor = 12;
    } else if (factor < 6) {
        factor = 6;
    }

    if (factor == 6) {
        __sx127x_lora_set_nb_trig_peaks(p_this, 5);
    } else {
        __sx127x_lora_set_nb_trig_peaks(p_this, 3);
    }

    modem_config2_reg = __sx127x_read_reg(p_this, REG_LR_MODEMCONFIG2);
    modem_config2_reg = (modem_config2_reg & RFLR_MODEMCONFIG2_SF_MASK) | (factor << 4);
    __sx127x_write_reg(p_this, REG_LR_MODEMCONFIG2, modem_config2_reg);
    p_this->p_config->p_lora_param->sf = factor;
}



static void __sx127x_lora_set_error_coding (radio_sx127x_dev_t *p_this, uint8_t value)
{
    uint8_t modem_config1_reg;
    
    modem_config1_reg = __sx127x_read_reg (p_this, REG_LR_MODEMCONFIG1);
    modem_config1_reg = (modem_config1_reg & RFLR_MODEMCONFIG1_CODINGRATE_MASK) | (value << 1);
    __sx127x_write_reg(p_this, REG_LR_MODEMCONFIG1, modem_config1_reg);
    p_this->p_config->p_lora_param->coderate = value;
}



static void __sx127x_lora_set_packet_crc_on (radio_sx127x_dev_t *p_this, bool enable)
{
    uint8_t modem_config2_reg;
    
    modem_config2_reg = __sx127x_read_reg(p_this, REG_LR_MODEMCONFIG2);
    modem_config2_reg = (modem_config2_reg & RFLR_MODEMCONFIG2_RXPAYLOADCRC_MASK) | (enable << 2);
    __sx127x_write_reg(p_this, REG_LR_MODEMCONFIG2, modem_config2_reg);
    p_this->p_config->p_lora_param->crc_on = enable;
}



static void __sx127x_lora_set_signal_bandwidth (radio_sx127x_dev_t *p_this, uint8_t bw)
{
    uint8_t modem_config1_reg;
    
    modem_config1_reg = __sx127x_read_reg(p_this, REG_LR_MODEMCONFIG1);
    modem_config1_reg = (modem_config1_reg & RFLR_MODEMCONFIG1_BW_MASK) | (bw << 4);
    __sx127x_write_reg(p_this, REG_LR_MODEMCONFIG1, modem_config1_reg);
    p_this->p_config->p_lora_param->bandwidth = bw;
}



static void __sx127x_lora_set_implicit_header_on (radio_sx127x_dev_t *p_this, bool enable)
{
    uint8_t modem_config1_reg;
    
    modem_config1_reg = __sx127x_read_reg(p_this, REG_LR_MODEMCONFIG1);
    modem_config1_reg = (modem_config1_reg & RFLR_MODEMCONFIG1_IMPLICITHEADER_MASK) | (enable);
    __sx127x_write_reg(p_this, REG_LR_MODEMCONFIG1, modem_config1_reg);
    p_this->p_config->p_lora_param->header_mode = enable;
}



static void __sx127x_lora_set_symb_timeout (radio_sx127x_dev_t *p_this, uint16_t value)
{
    uint8_t modem_config2_reg;
    uint8_t symb_timeout_lsb_reg;
    
    __sx127x_read_buffer(p_this, REG_LR_MODEMCONFIG2, &modem_config2_reg, 2);
    
    modem_config2_reg    = __sx127x_read_reg(p_this, REG_LR_MODEMCONFIG2);
    symb_timeout_lsb_reg = __sx127x_read_reg(p_this, REG_LR_SYMBTIMEOUTLSB);

    modem_config2_reg = (modem_config2_reg & RFLR_MODEMCONFIG2_SYMBTIMEOUTMSB_MASK) | ((value >> 8) & ~RFLR_MODEMCONFIG2_SYMBTIMEOUTMSB_MASK);
    symb_timeout_lsb_reg = value & 0xFF;
    
    __sx127x_write_reg(p_this, REG_LR_MODEMCONFIG2, modem_config2_reg);
    __sx127x_write_reg(p_this, REG_LR_SYMBTIMEOUTLSB, symb_timeout_lsb_reg);
}



static int __sx127x_lora_set_payload_length (radio_sx127x_dev_t *p_this, uint8_t value)
{
    __sx127x_write_reg(p_this, REG_LR_PAYLOADLENGTH, value);
    p_this->p_config->p_lora_param->payload_len = value;
    
    return RADIO_RET_OK;
}



static void __sx127x_lora_set_low_datarate_optimize (radio_sx127x_dev_t *p_this, bool enable)
{
    uint8_t modem_config3_reg;
    
    modem_config3_reg = __sx127x_read_reg(p_this, REG_LR_MODEMCONFIG3);
    modem_config3_reg = (modem_config3_reg & RFLR_MODEMCONFIG3_LOWDATARATEOPTIMIZE_MASK ) | (enable << 3);
    __sx127x_write_reg(p_this, REG_LR_MODEMCONFIG3, modem_config3_reg);
}



static void __sx127x_lora_set_pa_output (radio_sx127x_dev_t *p_this, uint8_t outputPin)
{
    uint8_t pa_config_reg;
    
    pa_config_reg = __sx127x_read_reg(p_this, REG_LR_PACONFIG);
    pa_config_reg = (pa_config_reg & RFLR_PACONFIG_PASELECT_MASK) | outputPin;
    __sx127x_write_reg(p_this, REG_LR_PACONFIG, pa_config_reg);
}



static void __sx127x_lora_set_pa_20dbm (radio_sx127x_dev_t *p_this, bool enale )
{
    uint8_t pa_dac_reg;
    uint8_t pa_config_reg;
    
    pa_dac_reg    = __sx127x_read_reg(p_this, REG_LR_PADAC);
    pa_config_reg = __sx127x_read_reg(p_this, REG_LR_PACONFIG);

    if ((pa_config_reg & RFLR_PACONFIG_PASELECT_PABOOST) == RFLR_PACONFIG_PASELECT_PABOOST) {
        if (enale == true) {
            pa_dac_reg = 0x87;
        }
    } else {
        pa_dac_reg = 0x84;
    }
    __sx127x_write_reg(p_this, REG_LR_PADAC, pa_dac_reg);  /*PA_BOOST*/
}



static uint8_t __sx127x_lora_set_rf_power (radio_sx127x_dev_t *p_this, int8_t power)
{
    uint8_t pa_dac_reg;
    uint8_t pa_config_reg;
    
    pa_config_reg = __sx127x_read_reg(p_this, REG_LR_PACONFIG);
    pa_dac_reg    = __sx127x_read_reg(p_this, REG_LR_PADAC);

    if ((pa_config_reg & RFLR_PACONFIG_PASELECT_PABOOST ) == RFLR_PACONFIG_PASELECT_PABOOST) {
        if ((pa_dac_reg & 0x87) == 0x87) {
            if (power < 5) {
                power = 5;
            }
            if (power > 20) {
                power = 20;
            }
            pa_config_reg = (pa_config_reg & RFLR_PACONFIG_MAX_POWER_MASK) | 0x70;
            pa_config_reg = (pa_config_reg & RFLR_PACONFIG_OUTPUTPOWER_MASK) | (uint8_t)((uint16_t)(power - 5) & 0x0F);
        } else {
            if (power < 2) {
                power = 2;
            }
            
            if (power > 17) {
                power = 17;
            }
            pa_config_reg = (pa_config_reg & RFLR_PACONFIG_MAX_POWER_MASK) | 0x70;
            pa_config_reg = (pa_config_reg & RFLR_PACONFIG_OUTPUTPOWER_MASK) | (uint8_t)((uint16_t)(power - 2) & 0x0F);
        }
    } else {
        if (power < -1) {
            power = -1;
        }
        
        if (power > 14) {
            power = 14;
        }
        pa_config_reg = (pa_config_reg & RFLR_PACONFIG_MAX_POWER_MASK) | 0x70;
        pa_config_reg = (pa_config_reg & RFLR_PACONFIG_OUTPUTPOWER_MASK) | (uint8_t)((uint16_t)(power + 1) & 0x0F);
    }
    __sx127x_write_reg(p_this, REG_LR_PACONFIG, pa_config_reg);
    p_this->p_config->p_lora_param->power = power;
    
    return power;
}



static int __sx127x_power_set_adapter (radio_sx127x_dev_t *p_this, uint8_t level, float *p_power)
{
    float power[] = {-1, 2, 5, 8, 11, 14, 17, 20};

    *p_power = __sx127x_lora_set_rf_power(p_this, power[level]);
    
    return RADIO_RET_OK;
}
    
    
    
static int __sx127x_lora_set_preamble_length (radio_sx127x_dev_t *p_this, uint16_t value)
{
    uint8_t preamble[2];
    
    __sx127x_read_buffer(p_this, REG_LR_PREAMBLEMSB, preamble, 2);

    preamble[0] = (value >> 8) & 0x00FF;
    preamble[1] = value & 0xFF;
    __sx127x_write_buffer(p_this, REG_LR_PREAMBLEMSB, preamble, 2);
    
    return RADIO_RET_OK;
}



static int __sx127x_lora_get_preamble_length (radio_sx127x_dev_t *p_this, uint16_t *p_len)
{
    uint8_t preamble[2];
    
    __sx127x_read_buffer(p_this, REG_LR_PREAMBLEMSB, preamble, 2);
    *p_len = ((preamble[0] & 0x00FF) << 8) | preamble[1];
    
    return RADIO_RET_OK;
}



static void __sx127x_lora_init (radio_sx127x_dev_t *p_this)
{
    uint8_t i;

    SM_STATE_SET(p_this, RFLR_STATE_IDLE);

    __sx127x_lora_set_defaults(p_this);

    __sx127x_write_reg(p_this, REG_LR_LNA, RFLR_LNA_GAIN_G1);

    // set the RF settings
    __sx127x_lora_set_rf_frequency(p_this, p_this->p_config->p_lora_param->frequency);
    __sx127x_lora_set_spreading_factor(p_this, p_this->p_config->p_lora_param->sf);
    __sx127x_lora_set_error_coding(p_this, p_this->p_config->p_lora_param->coderate);  
    __sx127x_lora_set_packet_crc_on(p_this, p_this->p_config->p_lora_param->crc_on);
    __sx127x_lora_set_signal_bandwidth(p_this, p_this->p_config->p_lora_param->bandwidth);

    __sx127x_lora_set_implicit_header_on(p_this, p_this->p_config->p_lora_param->header_mode);
    __sx127x_lora_set_symb_timeout(p_this, 0x3FF);
    __sx127x_lora_set_payload_length(p_this, p_this->p_config->p_lora_param->payload_len);
    __sx127x_lora_set_low_datarate_optimize(p_this, true);
    __sx127x_lora_set_preamble_length(p_this, p_this->p_config->p_lora_param->preamble_len); /* TODO:test cad rx timout */

    if (p_this->p_config->p_lora_param->frequency > 860000000) {
        __sx127x_lora_set_pa_output(p_this, RFLR_PACONFIG_PASELECT_RFO);
        __sx127x_lora_set_pa_20dbm(p_this, false);
        p_this->p_config->p_lora_param->power = 14;
        __sx127x_lora_set_rf_power(p_this, p_this->p_config->p_lora_param->power);
    } else {
        __sx127x_write_reg(p_this, REG_LR_OCP, 0x3f);    /* ĬϲPAڵΪ100mA,20dBmʱҪ120mA,Ա0x0b*/
        __sx127x_lora_set_pa_output(p_this, RFLR_PACONFIG_PASELECT_PABOOST);
        __sx127x_lora_set_pa_20dbm(p_this, true);
        __sx127x_lora_set_rf_power(p_this, p_this->p_config->p_lora_param->power);
    }

    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_STANDBY);

//    __sx127x_read_buffer(p_this, REG_LR_OPMODE, SX1276Regs + 1, 0x70 - 1);
//
//    for (i = 0; i < 0x70; i++) {
//        printf("0x%x 0x%x\n", i, SX1276Regs[i]);
//    }
}



static int __sx127x_mode_set (radio_sx127x_dev_t *p_this, uint8_t mode)
{
    int ret = RADIO_RET_OK;

    switch (mode) {
    case SLEEP_MODE:
        ret = __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_SLEEP);
        break;

    case STANDBY_MODE:
        ret = __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_STANDBY);
        break;
        
    case RX_MODE:
        ret = __sx127x_lora_rx_state_enter(p_this);
        break;
    
    case TX_MODE:
        ret = __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_TRANSMITTER);
        break;

    default:
        break;
    }

    return ret;
}



static int __sx127x_state_get (radio_sx127x_dev_t *p_this, radio_state_t *p_state)
{
    radio_state_t state;

    switch (SM_STATE_GET(p_this)) {
    case RFLR_STATE_TX_RUNNING:
        *p_state = TX_RUNING_ST;
        break;

    case RFLR_STATE_RX_RUNNING:
        *p_state = RX_RUNING_ST;
        break;

    case RFLR_STATE_CAD_RUNNING:
        *p_state = CAD_ST;
        break;

    case RFLR_STATE_IDLE:
        *p_state = IDLE_ST;
        break;

    default:
        break;
    }

    return RADIO_RET_OK;
}
    
    
    
static int __sx127x_lora_tx_data (radio_sx127x_dev_t *p_this, uint8_t *pbuf, uint8_t size)
{
    uint8_t irq_flags_mask_reg;
    uint8_t hop_period_reg;
    uint8_t hop_channel_reg;
    
    if ((NULL == pbuf) || (0 == size)) {
        return RADIO_RET_PARAM_INVALID;
    }
    
    if (size > 255) {
        return RADIO_RET_CFG_OUT_OF_RANGE;
    }

    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_STANDBY);
    if (p_this->p_config->p_lora_param->freq_hop_on == true) {
        irq_flags_mask_reg = RFLR_IRQFLAGS_RXTIMEOUT |
                             RFLR_IRQFLAGS_RXDONE |
                             RFLR_IRQFLAGS_PAYLOADCRCERROR |
                             RFLR_IRQFLAGS_VALIDHEADER |
                             //RFLR_IRQFLAGS_TXDONE |
                             RFLR_IRQFLAGS_CADDONE |
                             //RFLR_IRQFLAGS_FHSSCHANGEDCHANNEL |
                             RFLR_IRQFLAGS_CADDETECTED;
        
        hop_period_reg = p_this->p_config->p_lora_param->hop_period;

        hop_channel_reg = __sx127x_read_reg(p_this, REG_LR_HOPCHANNEL) & RFLR_HOPCHANNEL_CHANNEL_MASK;
        __sx127x_lora_set_rf_frequency(p_this, p_this->p_config->p_lora_param->p_hop_tab[hop_channel_reg % 
                                                                    p_this->p_config->p_lora_param->hop_tab_len]);
    } else {
        irq_flags_mask_reg = RFLR_IRQFLAGS_RXTIMEOUT |
                             RFLR_IRQFLAGS_RXDONE |
                             RFLR_IRQFLAGS_PAYLOADCRCERROR |
                             RFLR_IRQFLAGS_VALIDHEADER |
                             //RFLR_IRQFLAGS_TXDONE |                /*  ֻ򿪷ж      */
                             RFLR_IRQFLAGS_CADDONE |
                             RFLR_IRQFLAGS_FHSSCHANGEDCHANNEL |
                             RFLR_IRQFLAGS_CADDETECTED;
        hop_period_reg = 0;
    }
    __sx127x_write_reg(p_this, REG_LR_HOPPERIOD, hop_period_reg);
    __sx127x_write_reg(p_this, REG_LR_IRQFLAGSMASK, irq_flags_mask_reg);

    __sx127x_write_reg(p_this,  REG_LR_PAYLOADLENGTH, size); /* implicitģʽ(ʽͷ),дFIFO,0x80*/

    __sx127x_write_reg(p_this, REG_LR_FIFOTXBASEADDR, 0x00);    /* д뷢͵׵ַ */

    __sx127x_write_reg(p_this, REG_LR_FIFOADDRPTR, 0);       /* дFIFOķʵַ */
    
    __sx127x_write_fifo(p_this, pbuf, size);                 /* дҪ͵ */
    /* DIO0:TxDone   DIO1:Fhss */
    __sx127x_write_reg(p_this, 
                       REG_LR_DIOMAPPING1, 
                       RFLR_DIOMAPPING1_DIO0_01 | 
                       RFLR_DIOMAPPING1_DIO1_11 |
                       RFLR_DIOMAPPING1_DIO2_01 | 
                       RFLR_DIOMAPPING1_DIO3_11);
    
    __sx127x_write_reg(p_this, 
                       REG_LR_DIOMAPPING2, 
                       RFLR_DIOMAPPING2_DIO4_00 | 
                       RFLR_DIOMAPPING2_DIO5_11);

    SM_STATE_SET(p_this, RFLR_STATE_TX_RUNNING);
    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_TRANSMITTER);                /* 뷢״̬ */

    while (__sx127x_dio0_level_read(p_this) == 0);
    
    __sx127x_write_reg(p_this, REG_LR_IRQFLAGS, RFLR_IRQFLAGS_TXDONE);         /* ж */
    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_STANDBY);
    
    return RADIO_RET_OK;
}



static int8_t __get_lora_pkt_snr (radio_sx127x_dev_t *p_this)
{
    int8_t  pkt_snr;
    uint8_t snr_reg_val;

    /* ȡһݵ */
    snr_reg_val = __sx127x_read_reg(p_this, REG_LR_PKTSNRVALUE);     /*  */
    if (snr_reg_val & 0x80) {                                        /* С0 */
        pkt_snr = ((~snr_reg_val + 1) & 0xFF) >> 2;                  /* Invert and divide by 4 */
        pkt_snr = -pkt_snr;
    } else {                                                         /* ȴ0 */
        pkt_snr = (snr_reg_val & 0xFF) >> 2;                         /* Divide by 4 */
    }
    
    return pkt_snr;
}



static int16_t __get_lora_pkt_rssi (radio_sx127x_dev_t *p_this)
{  
    uint8_t rssi_reg;
    int16_t rssi_val;
    
    rssi_reg = __sx127x_read_reg(p_this, REG_LR_PKTRSSIVALUE);
    if (__get_lora_pkt_snr(p_this) < 0) {
        if (p_this->p_config->p_lora_param->frequency > __LORA_RF_MID_BAND_THRESH) {
            rssi_val = __LORA_RSSI_OFFSET_HF + rssi_reg + (rssi_reg >> 4) + p_this->pkt_info.snr;
        } else {
            rssi_val = __LORA_RSSI_OFFSET_LF + rssi_reg + (rssi_reg >> 4) + p_this->pkt_info.snr;
        }
    } else {
        if (p_this->p_config->p_lora_param->frequency > __LORA_RF_MID_BAND_THRESH) {
            rssi_val = __LORA_RSSI_OFFSET_HF + rssi_reg + (rssi_reg >> 4);
        } else {
            rssi_val = __LORA_RSSI_OFFSET_LF + rssi_reg + (rssi_reg >> 4);
        }
    }
    
    return rssi_val;
}



static int __cad_detect_state_get (radio_sx127x_dev_t *p_this)
{
    uint8_t reg_val;
    
    reg_val = __sx127x_read_reg(p_this, REG_LR_IRQFLAGS);
    if (reg_val & RFLR_IRQFLAGS_CADDONE) {
        CAD_STATE_SET(p_this, CAD_STATE_DONE);
        if (p_this->p_config->p_events->pfn_cad_done_cb != NULL) {
            p_this->p_config->p_events->pfn_cad_done_cb(true);
        }

        /* doneж */
        __sx127x_write_reg(p_this, REG_LR_IRQFLAGS, RFLR_IRQFLAGS_CADDONE); 

        /* Ƿcad_detectж */
        reg_val = __sx127x_read_reg(p_this, REG_LR_IRQFLAGS);
        if (reg_val & RFLR_IRQFLAGS_CADDETECTED) {
            CAD_STATE_SET(p_this, CAD_STATE_DETECT);
            /* detectж */
            __sx127x_write_reg(p_this, REG_LR_IRQFLAGS, RFLR_IRQFLAGS_CADDONE); 

            return LORA_RET_CODE_OK;
        } else {
            // The device goes in Standby Mode automatically    
            SM_STATE_SET(p_this, RFLR_STATE_IDLE);
            CAD_STATE_SET(p_this, CAD_STATE_FINISH);
            return LORA_RET_CODE_NO_CAD_DETECT_IRQ;
        }
    }   

    return LORA_RET_CODE_NO_CAD_DONE_IRQ;
}



static int __sx127x_lora_read_cur_rssi (radio_sx127x_dev_t *p_this, int16_t *p_rssi)
{
    uint8_t rssi_reg;
    
    /* ˴ֻܲ뵽߽ӿڴĹʣźźźǿȲһ
     * ΪźǿȿͨSNRԭ֮µźǿ 
     */ 
    rssi_reg = __sx127x_read_reg(p_this, REG_LR_RSSIVALUE);
    *p_rssi = RssiOffsetHF[p_this->p_config->p_lora_param->bandwidth] + rssi_reg;
    
    return RADIO_RET_OK;
}



static int __sx127x_lora_rx_data_read (radio_sx127x_dev_t *p_this, uint8_t *p_buf, uint8_t *p_size)
{
    uint8_t irq_flags_reg;
    uint8_t fifo_rx_cur_addr;
    
    if (__sx127x_dio0_level_read(p_this) == 1) {                                     /* RxDone                     */
        __sx127x_write_reg(p_this,  REG_LR_IRQFLAGS, RFLR_IRQFLAGS_RXDONE);          /* ִджϱ־ */
    }

    irq_flags_reg = __sx127x_read_reg(p_this, REG_LR_IRQFLAGS);
    if ((irq_flags_reg & RFLR_IRQFLAGS_PAYLOADCRCERROR) == RFLR_IRQFLAGS_PAYLOADCRCERROR) {  /* CRC */
        __sx127x_write_reg(p_this, REG_LR_IRQFLAGS, RFLR_IRQFLAGS_PAYLOADCRCERROR);

        if (p_this->p_config->p_events != NULL) {
            if (p_this->p_config->p_events->pfn_rx_error_cb != NULL) {
                p_this->p_config->p_events->pfn_rx_error_cb();
            }
        }

        *p_size = 0;
        return LORA_RET_RECV_CRC_ERR;
    }

    p_this->pkt_info.snr  = __get_lora_pkt_snr(p_this);             /* ȡݰ   */
    p_this->pkt_info.rssi = __get_lora_pkt_rssi(p_this);            /* ȡݰźǿ */

    /* ȡ */
    fifo_rx_cur_addr = __sx127x_read_reg(p_this, REG_LR_FIFORXCURRENTADDR);  /* ȡһָ */
    if (p_this->p_config->p_lora_param->header_mode == IMPLICIT_HEADER) {
        __sx127x_write_reg(p_this, REG_LR_FIFOADDRPTR, fifo_rx_cur_addr);    /* дFIFOķʵַ */
        __sx127x_read_fifo(p_this, p_buf, p_this->p_config->p_lora_param->payload_len);
        *p_size = p_this->p_config->p_lora_param->payload_len;
    } else {
        *p_size = __sx127x_read_reg(p_this, REG_LR_NBRXBYTES);         /* ȡ */
        __sx127x_write_reg(p_this,  REG_LR_FIFOADDRPTR, 0);  /* дFIFOķʵַ */
        __sx127x_read_fifo(p_this, p_buf, *p_size);
    }
    
    if (p_this->p_config->p_events != NULL) {
        if (p_this->p_config->p_events->pfn_rx_done_cb != NULL) {
            p_this->p_config->p_events->pfn_rx_done_cb(p_buf, 
                                                       *p_size, 
                                                       p_this->pkt_info.rssi, 
                                                       p_this->pkt_info.snr);
        }
    }
    
    if (p_this->p_config->p_lora_param->rx_mode == CONTINUOUS_RECV_MODE) {
        SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
    } else {
        SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    }

    return LORA_RET_CODE_OK;
}



static int __sx127x_cad_start (radio_sx127x_dev_t *p_this)
{
    uint8_t reg_val;
    
    /* ǰCADδн¿CAD */
    if (CAD_STATE_GET(p_this) & CAD_STATE_START) {
        if (!(CAD_STATE_GET(p_this) & CAD_STATE_FINISH)) {
            return LORA_RET_CAD_RUNNING;
        }
    }
    
    if (p_this->p_config->p_lora_param->freq_hop_on == 1) {
        __sx127x_lora_set_rf_frequency(p_this, p_this->p_config->p_lora_param->p_hop_tab[0]);
    }

    /* ʼCAD״̬ */
    CAD_STATE_CLR(p_this);

    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_STANDBY);
    
    __sx127x_write_reg(p_this, 
                       REG_LR_IRQFLAGSMASK, 
                       RFLR_IRQFLAGS_RXTIMEOUT |
                       RFLR_IRQFLAGS_RXDONE |
                       RFLR_IRQFLAGS_PAYLOADCRCERROR |
                       RFLR_IRQFLAGS_VALIDHEADER |
                       RFLR_IRQFLAGS_TXDONE |
                       //RFLR_IRQFLAGS_CADDONE |
                       RFLR_IRQFLAGS_FHSSCHANGEDCHANNEL //|
                       /* RFLR_IRQFLAGS_CADDETECTED*/ );
    
    __sx127x_write_reg(p_this, 
                       REG_LR_DIOMAPPING1,
                       RFLR_DIOMAPPING1_DIO0_10 |  /* CAD Done */
                       RFLR_DIOMAPPING1_DIO1_00 |  /* RxTimeout */
                       RFLR_DIOMAPPING1_DIO2_00 |  /* FhssChangeChannel */
                       RFLR_DIOMAPPING1_DIO3_00);  /* CadDone */
     
    __sx127x_write_reg(p_this, 
                       REG_LR_DIOMAPPING2,
                       RFLR_DIOMAPPING2_DIO4_00 |  /* CAD Detected */
                       RFLR_DIOMAPPING2_DIO5_00);  /* ModeReady */
     
    SM_STATE_SET(p_this, RFLR_STATE_CAD_RUNNING);
    CAD_STATE_SET(p_this, CAD_STATE_START);
    __sx127x_lora_set_op_mode(p_this, RFLR_OPMODE_CAD); 
    
    return RADIO_RET_OK;
}
    
    

static uint32_t __sx127x_time_on_air_get (radio_sx127x_dev_t *p_this,
                                          uint8_t             pkt_len)
{
    uint32_t len, bw, dr, cr, preamble_len;
    uint64_t ts, tpreamble, tpayload, tmp;    /* unit: 1e-6 */
    uint32_t air_time = 0;

    len          = pkt_len;
    dr           = p_this->p_config->p_lora_param->sf;
    cr           = p_this->p_config->p_lora_param->coderate;
    preamble_len = p_this->p_config->p_lora_param->preamble_len;

    // REMARK: When using LoRa modem only bandwidths 125, 250 and 500 kHz are supported
    switch (p_this->p_config->p_lora_param->bandwidth) {
        case 7: // 125 kHz
            bw = 125 * 1000;
            break;

        case 8: // 250 kHz
            bw = 250 * 1000;
            break;

        case 9: // 500 kHz
            bw = 500 * 1000;
            break;

        default:
            return 0;
    }

    // Symbol rate : time for one symbol (secs)
    ts = ((uint64_t)(1 << dr)) * (uint64_t)1000000 / bw;

    // time of preamble
    tpreamble = (preamble_len * (uint64_t)1000000 + (uint64_t)(4.25 * 1000000)) * ts / (uint64_t)1000000;

    // Symbol length of payload and time
    tmp = (8 * len - 4 * dr + 28 +
           16 * (p_this->p_config->p_lora_param->crc_on ? 1 : 0) -
           (p_this->p_config->p_lora_param->header_mode ? 20 : 0)) * (uint64_t)(1000000)
        / (4 * (dr - ((0 > 0) ? 2 : 0)));

    /* ceil */
    if (tmp > 0) {
        if (tmp % (uint64_t)1000000 < 10000) {
            tmp = ((tmp / (uint64_t)1000000) * (cr + 4)) * (uint64_t)1000000;
        } else {
            tmp = ((1 + (tmp / (uint64_t)1000000)) * (cr + 4)) * (uint64_t)1000000;
        }
    } else {
        tmp = 0;
    }

    tpayload = (8 * (uint64_t)1000000 + tmp) * ts / (uint64_t)1000000;

    air_time = 1 + (tpreamble + tpayload) / (uint64_t)1000;

    // Time on air, ms
    return air_time;
}



static int __sx127x_ioctl (radio_sx127x_dev_t *p_this, uint16_t cmd, void *p_arg)
{
    uint8_t reg;
    int     ret = RADIO_RET_OK;
    
    switch (cmd) {
        case CAD_START:
            ret = __sx127x_cad_start(p_this);
            break;
        
        default:
            break;
    }
    
    return ret;
}



void __sx127x_lora_dio0_irq_func (radio_sx127x_dev_t *p_this)
{
    switch (SM_STATE_GET(p_this)) {
    case RFLR_STATE_CAD_RUNNING:
        if (__cad_detect_state_get(p_this) == LORA_RET_CODE_OK) {
            /* лģʽ */
            __sx127x_lora_rx_state_enter(p_this);
            SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
        } else {
            CAD_STATE_SET(p_this, CAD_STATE_FINISH);
        }
        break;
        
    case RFLR_STATE_RX_RUNNING:
        __sx127x_lora_rx_data_read(p_this, p_this->pkt_info.pkt_buffer, &p_this->pkt_info.pkt_len);
        if (CAD_STATE_GET(p_this) & CAD_STATE_START) {
            CAD_STATE_SET(p_this, CAD_STATE_FINISH);
        }

        break;

    default:
        break;
    }
}



void __sx127x_lora_dio1_irq_func (radio_sx127x_dev_t *p_this)
{
    switch (SM_STATE_GET(p_this)) {
    case RFLR_STATE_RX_RUNNING:
        // Clear Irq
        __sx127x_write_reg(p_this,
                           REG_LR_IRQFLAGS,
                           RFLR_IRQFLAGS_RXTIMEOUT);

        SM_STATE_SET(p_this, RFLR_STATE_IDLE);
        /* пCADΪ״̬ */
        if (CAD_STATE_GET(p_this) & CAD_STATE_START) {
            CAD_STATE_SET(p_this, CAD_STATE_FINISH);
        }

        if ((p_this->p_config->p_events != NULL) &&
            (p_this->p_config->p_events->pfn_rx_timeout_cb != NULL)) {
            p_this->p_config->p_events->pfn_rx_timeout_cb();
        }
        break;

    default:
        break;
    }
}



void __sx127x_lora_dio2_irq_func (radio_sx127x_dev_t *p_this)
{
    uint8_t hop_channel_reg;
    
    switch (SM_STATE_GET(p_this)) {
    case RFLR_STATE_RX_RUNNING:
    case RFLR_STATE_TX_RUNNING:
        if (p_this->p_config->p_lora_param->freq_hop_on == true ) {
            hop_channel_reg = __sx127x_read_reg(p_this, REG_LR_HOPCHANNEL) & RFLR_HOPCHANNEL_CHANNEL_MASK;
            __sx127x_lora_set_rf_frequency(p_this, p_this->p_config->p_lora_param->p_hop_tab[hop_channel_reg % 
                                                                    p_this->p_config->p_lora_param->hop_tab_len]);

            if (p_this->p_config->p_events != NULL) {
                if (p_this->p_config->p_events->pfn_fhss_change_channel_cb != NULL) {
                    p_this->p_config->p_events->pfn_fhss_change_channel_cb(hop_channel_reg & RFLR_HOPCHANNEL_CHANNEL_MASK);
                }
            }
        }

        // Clear Irq
        __sx127x_write_reg(p_this, REG_LR_IRQFLAGS, RFLR_IRQFLAGS_FHSSCHANGEDCHANNEL);
        break;

    default:
        break;
    }
}



static int __sx127x_pkt_buf_read (radio_sx127x_dev_t *p_this, uint8_t *p_buf, uint8_t *p_size)
{
    if (p_this->pkt_info.pkt_len > 0) {
        memcpy(p_buf, p_this->pkt_info.pkt_buffer, p_this->pkt_info.pkt_len);
        *p_size = p_this->pkt_info.pkt_len;
        memset(p_this->pkt_info.pkt_buffer, 0, AM_LORA_RX_BUFFER_SIZE);
        p_this->pkt_info.pkt_len = 0;
        return RADIO_RET_OK;
    }
    
    return RADIO_RET_NOT_RECV_PKT;
}



static int spi_check (radio_sx127x_dev_t *p_this)
{
    uint8_t reg = 0x35;
    
    __sx127x_write_reg(p_this, REG_LR_FRFMID, reg);
    if (reg != __sx127x_read_reg(p_this, REG_LR_FRFMID)) {
        return -1;
    }
    
    return 0;
}



static int __sx127x_sync_id_set (radio_sx127x_dev_t *p_this, uint8_t *p_id, uint8_t len)
{
    return RADIO_RET_MODULE_NOT_SUPPORT;
}
    
    

static int __sx127x_sync_id_get (radio_sx127x_dev_t *p_this, uint8_t *p_id, uint8_t *p_len)
{
    return RADIO_RET_MODULE_NOT_SUPPORT;
}



static const struct radio_drv_funcs __g_sx127x_drv_funcs[] = {
    __sx127x_reset,
    __sx127x_sync_id_set, 
    __sx127x_sync_id_get,                 
    __sx127x_lora_set_preamble_length,   
    __sx127x_lora_get_preamble_length,    
    __sx127x_power_set_adapter,            
    __sx127x_lora_set_rf_frequency,   
    __sx127x_mode_set,
    __sx127x_state_get,                    /* TODO:need test */
    __sx127x_lora_read_cur_rssi,           
    __sx127x_lora_tx_data,
    __sx127x_pkt_buf_read, 
    __sx127x_ioctl,
    __sx127x_lora_dio0_irq_func,
    __sx127x_lora_dio1_irq_func,
    __sx127x_lora_dio2_irq_func,
    NULL,
    NULL,
};



radio_handle_t radio_sx127x_init (radio_sx127x_dev_t      *p_sx127x,
                                  sx127x_spi_funcs_t      *p_spi,
                                  sx127x_gpio_funcs_t     *p_gpio,
                                  sx127x_timer_funcs_t    *p_delay,
                                  sx127x_config_t         *p_config)
{    
    if ((p_sx127x == NULL) || (p_spi == NULL)) {
        return NULL;
    }
    
    p_sx127x->p_spi        = p_spi;
    p_sx127x->p_gpio       = p_gpio;
    p_sx127x->p_delay      = p_delay;
    p_sx127x->p_config     = p_config;
    p_sx127x->isa.p_funcs  = __g_sx127x_drv_funcs;
    p_sx127x->isa.p_drv    = p_sx127x;
    
    CAD_STATE_CLR(p_sx127x);
    SM_STATE_SET(p_sx127x, RFLR_STATE_IDLE);

    if (spi_check(p_sx127x) != 0) {
        return NULL;
    }
    
    __sx127x_reset(p_sx127x);

    __sx127x_set_lora_on(p_sx127x, true);

    __sx127x_lora_init(p_sx127x);

    return &p_sx127x->isa;
}
